@@ -1,11 +1,11 @@
 ;;; csv-mode.el --- Major mode for editing comma/char separated values  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2003, 2004, 2012-2020  Free Software Foundation, Inc
+;; Copyright (C) 2003, 2004, 2012-2022 Free Software Foundation, Inc
 
 ;; Author: "Francis J. Wright" <F.J.Wright@qmul.ac.uk>
 ;; Maintainer: emacs-devel@gnu.org
-;; Version: 1.19
-;; Package-Requires: ((emacs "24.1") (cl-lib "0.5"))
+;; Version: 1.21
+;; Package-Requires: ((emacs "27.1") (cl-lib "0.5"))
 ;; Keywords: convenience
 
 ;; This package is free software; you can redistribute it and/or modify
@@ -54,6 +54,15 @@
 ;; - C-c C-t (`csv-transpose') interchanges rows and columns.  For
 ;;   details, see the documentation for the individual commands.
 
+;; - `csv-set-separator' sets the CSV separator of the current buffer,
+;;   while `csv-guess-set-separator' guesses and sets the separator
+;;   based on the current buffer's contents.
+;;   `csv-guess-set-separator' can be useful to add to the mode hook
+;;   to have CSV mode guess and set the separator automatically when
+;;   visiting a buffer:
+;;
+;;     (add-hook 'csv-mode-hook 'csv-guess-set-separator)
+
 ;; CSV mode can recognize fields separated by any of several single
 ;; characters, specified by the value of the customizable user option
 ;; `csv-separators'.  CSV data fields can be delimited by quote
@@ -119,7 +128,9 @@
 
 ;;; Code:
 
-(eval-when-compile (require 'cl-lib))
+(eval-when-compile
+  (require 'cl-lib)
+  (require 'subr-x))
 
 (defgroup CSV nil
   "Major mode for editing files of comma-separated value type."
@@ -163,12 +174,14 @@ session.  Use `customize-set-variable' instead if that is required."
                      (error "%S is already a quote" x)))
 	       value)
 	 (custom-set-default variable value)
-	 (setq csv-separator-chars (mapcar #'string-to-char value)
-	       csv--skip-chars (apply #'concat "^\n" csv-separators)
-	       csv-separator-regexp (apply #'concat `("[" ,@value "]"))
-	       csv-font-lock-keywords
-	       ;; NB: csv-separator-face variable evaluates to itself.
-	       `((,csv-separator-regexp (0 'csv-separator-face))))))
+         (setq csv-separator-chars (mapcar #'string-to-char value))
+         (setq csv--skip-chars
+               (apply #'concat "^\n"
+                      (mapcar (lambda (s) (concat "\\" s)) value)))
+         (setq csv-separator-regexp (regexp-opt value))
+         (setq csv-font-lock-keywords
+               ;; NB: csv-separator-face variable evaluates to itself.
+               `((,csv-separator-regexp (0 'csv-separator-face))))))
 
 (defcustom csv-field-quotes '("\"")
   "Field quotes: a list of *single-character* strings.
@@ -213,7 +226,7 @@ FIELD-QUOTES should be a list of single-character strings."
 (defvar csv-comment-start nil
   "String that starts a comment line, or nil if no comment syntax.
 Such comment lines are ignored by CSV mode commands.
-This variable is buffer local\; its default value is that of
+This variable is buffer local; its default value is that of
 `csv-comment-start-default'.  It is set by the function
 `csv-set-comment-start' -- do not set it directly!")
 
@@ -348,7 +361,7 @@ It must be either a string or nil."
    (list (edit-and-eval-command
 	  "Comment start (string or nil): " csv-comment-start)))
   ;; Paragraph means a group of contiguous records:
-  (set (make-local-variable 'paragraph-separate) "[:space:]*$") ; White space.
+  (set (make-local-variable 'paragraph-separate) "[[:space:]]*$") ; White space.
   (set (make-local-variable 'paragraph-start) "\n");Must include \n explicitly!
   ;; Remove old comment-start/end if available
   (with-syntax-table text-mode-syntax-table
@@ -368,6 +381,23 @@ It must be either a string or nil."
     (modify-syntax-entry ?\n ">" csv-mode-syntax-table))
   (setq csv-comment-start string))
 
+(defvar csv--set-separator-history nil)
+
+(defun csv-set-separator (sep)
+  "Set the CSV separator in the current buffer to SEP."
+  (interactive (list (read-char-from-minibuffer
+                      "Separator: " nil 'csv--set-separator-history)))
+  (when (and (boundp 'csv-field-quotes)
+             (member (string sep) csv-field-quotes))
+    (error "%c is already a quote" sep))
+  (setq-local csv-separators (list (string sep)))
+  (setq-local csv-separator-chars (list sep))
+  (setq-local csv--skip-chars (format "^\n\\%c" sep))
+  (setq-local csv-separator-regexp (regexp-quote (string sep)))
+  (setq-local csv-font-lock-keywords
+              `((,csv-separator-regexp (0 'csv-separator-face))))
+  (font-lock-refresh-defaults))
+
 ;;;###autoload
 (add-to-list 'auto-mode-alist '("\\.[Cc][Ss][Vv]\\'" . csv-mode))
 
@@ -389,7 +419,7 @@ Usually they sort in order of ascending sort key.")
       (remove-from-invisibility-spec 'csv)
     (add-to-invisibility-spec 'csv))
   (message "Separators in aligned records will be %svisible \
-\(after re-aligning if soft\)"
+\(after re-aligning if soft)"
 	   (if (memq 'csv buffer-invisibility-spec) "in" ""))
   (redraw-frame (selected-frame)))
 
@@ -473,10 +503,10 @@ Assumes point is at beginning of line."
 Signal an error if the buffer is read-only.
 If TYPE is noarg then return a list (beg end).
 Otherwise, return a list (arg beg end), where arg is:
-  the raw prefix argument by default\;
-  a single field index if TYPE is single\;
+  the raw prefix argument by default;
+  a single field index if TYPE is single;
   a list of field indices or index ranges if TYPE is multiple.
-Field defaults to the current prefix arg\; if not set, prompt user.
+Field defaults to the current prefix arg; if not set, prompt user.
 
 A field index list consists of positive or negative integers or ranges,
 separated by any non-integer characters.  A range has the form m-n,
@@ -604,7 +634,7 @@ Field indices increase from 1 on the left or decrease from -1 on the right.
 A prefix argument specifies a single field, otherwise prompt for field index.
 Ignore blank and comment lines.  The variable `sort-fold-case'
 determines whether alphabetic case affects the sort order.
-When called non-interactively, FIELD is a single field index\;
+When called non-interactively, FIELD is a single field index;
 BEG and END specify the region to sort."
   ;; (interactive "*P\nr")
   (interactive (csv-interactive-args 'single))
@@ -624,7 +654,7 @@ Specified non-null field must contain a number in each line of the region,
 which may begin with \"0x\" or \"0\" for hexadecimal and octal values.
 Otherwise, the number is interpreted according to sort-numeric-base.
 Ignore blank and comment lines.
-When called non-interactively, FIELD is a single field index\;
+When called non-interactively, FIELD is a single field index;
 BEG and END specify the region to sort."
   ;; (interactive "*P\nr")
   (interactive (csv-interactive-args 'single))
@@ -968,14 +998,14 @@ Ignore blank and comment lines."
 (defun csv-yank-fields (field beg end)
   "Yank fields as the ARGth field of each line in the region.
 ARG may be arbitrarily large and records are extended as necessary.
-If not set, the region defaults to the CSV records around point\;
+If not set, the region defaults to the CSV records around point;
 if point is not in a CSV record then offer to yank as a new table.
 The fields yanked are those last killed by `csv-kill-fields'.
 Fields are separated by `csv-separators' and null fields are allowed anywhere.
 Field indices increase from 1 on the left or decrease from -1 on the right.
 A prefix argument specifies a single field, otherwise prompt for field index.
 Ignore blank and comment lines.  When called non-interactively, FIELD
-is a single field index\; BEG and END specify the region to process."
+is a single field index; BEG and END specify the region to process."
   ;; (interactive "*P\nr")
   (interactive (condition-case err
 		   (csv-interactive-args 'single)
@@ -1072,10 +1102,10 @@ Unalign first (see `csv-unalign-fields').  Ignore blank and comment lines.
 In hard-aligned records, separators become invisible whenever
 `buffer-invisibility-spec' is non-nil.  In soft-aligned records, make
 separators invisible if and only if `buffer-invisibility-spec' is
-non-nil when the records are aligned\; this can be changed only by
+non-nil when the records are aligned; this can be changed only by
 re-aligning.  \(Unaligning always makes separators visible.)
 
-When called non-interactively, use hard alignment if HARD is non-nil\;
+When called non-interactively, use hard alignment if HARD is non-nil;
 BEG and END specify the region to align.
 If there is no selected region, default to the whole buffer."
   (interactive (cons current-prefix-arg
@@ -1205,7 +1235,7 @@ Undo soft alignment introduced by `csv-align-fields'.  If invoked with
 an argument then also remove all spaces and tabs around separators.
 Also make all invisible separators visible again.
 Ignore blank and comment lines.  When called non-interactively, remove
-spaces and tabs if HARD non-nil\; BEG and END specify region to unalign.
+spaces and tabs if HARD non-nil; BEG and END specify region to unalign.
 If there is no selected region, default to the whole buffer."
   (interactive (cons current-prefix-arg
                      (if (use-region-p)
@@ -1356,8 +1386,12 @@ If there is already a header line, then unset the header line."
       (overlay-put csv--header-line 'modification-hooks
                    '(csv--header-flush)))
     (csv--header-flush)
+    ;; These are introduced in Emacs 29.
+    (unless (boundp 'header-line-indent)
+      (setq-local header-line-indent ""
+                  header-line-indent-width 0))
     (setq header-line-format
-          '(:eval (csv--header-string)))))
+          '("" header-line-indent (:eval (csv--header-string))))))
 
 (defun csv--header-flush (&rest _)
   ;; Force re-computation of the header-line.
@@ -1391,9 +1425,10 @@ If there is already a header line, then unset the header line."
                         (nexti (next-single-property-change i 'display str))
                         (newprop
                          `(space :align-to
-                                 ,(if (numberp x)
-                                      (- x (or csv--header-hscroll 0))
-                                    `(- ,x csv--header-hscroll)))))
+                                 (+ ,(if (numberp x)
+                                         (- x (or csv--header-hscroll 0))
+                                       `(- ,x csv--header-hscroll))
+                                    header-line-indent-width))))
                    (put-text-property i (or nexti (length str))
                                       'display newprop str)
                    (setq i nexti))))
@@ -1493,10 +1528,11 @@ setting works better)."
     (delete-overlay ol)))
 
 (defun csv--jit-unalign (beg end)
-  (remove-text-properties beg end
-                          '(display nil csv--jit nil invisible nil
-                            cursor-sensor-functions nil csv--revealed nil))
-  (remove-overlays beg end 'csv--jit t))
+  (with-silent-modifications
+    (remove-text-properties beg end
+                            '( display nil csv--jit nil invisible nil
+                               cursor-sensor-functions nil csv--revealed nil))
+    (remove-overlays beg end 'csv--jit t)))
 
 (defun csv--jit-flush (beg end)
   "Cause all the buffer (except for the BEG...END region) to be re-aligned."
@@ -1721,6 +1757,8 @@ setting works better)."
     (add-to-invisibility-spec '(csv-truncate . t))
     (kill-local-variable 'csv--jit-columns)
     (cursor-sensor-mode 1)
+    (when (fboundp 'header-line-indent-mode)
+      (header-line-indent-mode))
     (jit-lock-register #'csv--jit-align)
     (jit-lock-refontify))
    (t
@@ -1728,6 +1766,104 @@ setting works better)."
     (jit-lock-unregister #'csv--jit-align)
     (csv--jit-unalign (point-min) (point-max))))
   (csv--header-flush))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;;  Separator guessing
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(defvar csv--preferred-separators
+  '(?, ?\; ?\t)
+  "Preferred separator characters in case of a tied score.")
+
+(defun csv-guess-set-separator ()
+  "Guess and set the CSV separator of the current buffer.
+
+Add it to the mode hook to have CSV mode guess and set the
+separator automatically when visiting a buffer:
+
+  (add-hook \\='csv-mode-hook \\='csv-guess-set-separator)"
+  (interactive)
+  (let ((sep (csv-guess-separator
+              (buffer-substring-no-properties
+               (point-min)
+               ;; We're probably only going to look at the first 2048
+               ;; or so chars, but take more than we probably need to
+               ;; minimize the chance of breaking the input in the
+               ;; middle of a (long) row.
+               (min 8192 (point-max)))
+              2048)))
+    (when sep
+      (csv-set-separator sep))))
+
+(defun csv-guess-separator (text &optional cutoff)
+  "Return a guess of which character is the CSV separator in TEXT."
+  (let ((best-separator nil)
+        (best-score 0))
+    (dolist (candidate (csv--separator-candidates text cutoff))
+      (let ((candidate-score
+             (csv--separator-score candidate text cutoff)))
+        (when (or (> candidate-score best-score)
+                  (and (= candidate-score best-score)
+                       (member candidate csv--preferred-separators)))
+          (setq best-separator candidate)
+          (setq best-score candidate-score))))
+    best-separator))
+
+(defun csv--separator-candidates (text &optional cutoff)
+  "Return a list of candidate CSV separators in TEXT.
+When CUTOFF is passed, look only at the first CUTOFF number of characters."
+  (let ((chars (make-hash-table)))
+    (dolist (c (string-to-list
+                (if cutoff
+                    (substring text 0 (min cutoff (length text)))
+                  text)))
+      (when (and (not (gethash c chars))
+                 (or (= c ?\t)
+                     (and (not (member c '(?. ?/ ?\" ?')))
+                          (not (member (get-char-code-property c 'general-category)
+                                       '(Lu Ll Lt Lm Lo Nd Nl No Ps Pe Cc Co))))))
+        (puthash c t chars)))
+    (hash-table-keys chars)))
+
+(defun csv--separator-score (separator text &optional cutoff)
+  "Return a score on how likely SEPARATOR is a separator in TEXT.
+
+When CUTOFF is passed, stop the calculation at the next whole
+line after having read CUTOFF number of characters.
+
+The scoring is based on the idea that most CSV data is tabular,
+i.e. separators should appear equally often on each line.
+Furthermore, more commonly appearing characters are scored higher
+than those who appear less often.
+
+Adapted from the paper \"Wrangling Messy CSV Files by Detecting
+Row and Type Patterns\" by Gerrit J.J. van den Burg , Alfredo
+Nazábal, and Charles Sutton: https://arxiv.org/abs/1811.11242."
+  (let ((groups
+         (with-temp-buffer
+           (csv-set-separator separator)
+           (save-excursion
+             (insert text))
+           (let ((groups (make-hash-table))
+                 (chars-read 0))
+             (while (and (/= (point) (point-max))
+                         (or (not cutoff)
+                             (< chars-read cutoff)))
+               (let* ((lep (line-end-position))
+                      (nfields (length (csv--collect-fields lep))))
+                 (cl-incf (gethash nfields groups 0))
+                 (cl-incf chars-read (- lep (point)))
+                 (goto-char (+ lep 1))))
+             groups)))
+        (sum 0))
+    (maphash
+     (lambda (length num)
+       (cl-incf sum (* num (/ (- length 1) (float length)))))
+     groups)
+    (let ((unique-groups (hash-table-count groups)))
+      (if (= 0 unique-groups)
+          0
+        (/ sum unique-groups)))))
 
 ;;; TSV support
 
